@@ -6,7 +6,7 @@ JSON::Validator - Validate data against a JSON schema
 
 =head1 VERSION
 
-0.53
+0.54
 
 =head1 SYNOPSIS
 
@@ -81,21 +81,23 @@ use Mojo::JSON::Pointer;
 use Mojo::URL;
 use Mojo::Util;
 use B;
+use Cwd            ();
 use File::Basename ();
 use File::Spec;
 use Scalar::Util;
 
 use constant VALIDATE_HOSTNAME => eval 'require Data::Validate::Domain;1';
 use constant VALIDATE_IP       => eval 'require Data::Validate::IP;1';
-use constant IV_SIZE           => eval 'require Config;$Config::Config{ivsize}';
 
 use constant DEBUG => $ENV{JSON_VALIDATOR_DEBUG} || $ENV{SWAGGER2_DEBUG} || 0;
 use constant WARN_ON_MISSING_FORMAT => $ENV{JSON_VALIDATOR_WARN_ON_MISSING_FORMAT}
   || $ENV{SWAGGER2_WARN_ON_MISSING_FORMAT} ? 1 : 0;
 
-our $VERSION = '0.53';
+our $VERSION = '0.54';
 our @EXPORT_OK = qw( validate_json );
 
+my $HTTP_SCHEME_RE = qr{^https?:};
+
 sub E { bless {path => $_[0] || '/', message => $_[1]}, 'JSON::Validator::Error'; }
 sub S { Mojo::Util::md5_sum(Data::Dumper->new([@_])->Sortkeys(1)->Useqq(1)->Dump); }
 
@@ -148,47 +150,19 @@ Note! The modules mentioned below are optional.
 
 =over 4
 
-=item * byte
-
-A padded, base64-encoded string of bytes, encoded with a URL and filename safe
-alphabet. Defined by RFC4648.
-
-=item * date
-
-An RFC3339 date in the format YYYY-MM-DD
-
 =item * date-time
 
 An RFC3339 timestamp in UTC time. This is formatted as
 "YYYY-MM-DDThh:mm:ss.fffZ". The milliseconds portion (".fff") is optional
 
-=item * double
-
-Cannot test double values with higher precision then what
-the "number" type already provides.
-
 =item * email
 
 Validated against the RFC5322 spec.
 
-=item * float
-
-Will always be true if the input is a number, meaning there is no difference
-between  L</float> and L</double>. Patches are welcome.
-
 =item * hostname
 
 Will be validated using L<Data::Validate::Domain> if installed.
 
-=item * int32
-
-A signed 32 bit integer.
-
-=item * int64
-
-A signed 64 bit integer. Note: This check is only available if Perl is
-compiled to use 64 bit integers.
-
 =item * ipv4
 
 Will be validated using L<Data::Validate::IP> if installed or
@@ -224,22 +198,7 @@ has cache_dir => sub {
 
 has coerce => $ENV{JSON_VALIDATOR_COERCE_VALUES} || $ENV{SWAGGER_COERCE_VALUES} || 0;    # EXPERIMENTAL!
 
-has formats => sub {
-  +{
-    'byte'      => \&_is_byte_string,
-    'date'      => \&_is_date,
-    'date-time' => \&_is_date_time,
-    'double'    => sub {1},
-    'float'     => sub {1},
-    'email'     => \&_is_email,
-    'hostname'  => VALIDATE_HOSTNAME ? \&Data::Validate::Domain::is_domain : \&_is_domain,
-    'int32'     => sub { _is_number($_[0], 'l'); },
-    'int64'     => IV_SIZE >= 8 ? sub { _is_number($_[0], 'q'); } : sub {1},
-    'ipv4' => VALIDATE_IP ? \&Data::Validate::IP::is_ipv4 : \&_is_ipv4,
-    'ipv6' => VALIDATE_IP ? \&Data::Validate::IP::is_ipv6 : \&_is_ipv6,
-    'uri'  => \&_is_uri,
-  };
-};
+has formats => sub { shift->_build_formats };
 
 has ua => sub {
   require Mojo::UserAgent;
@@ -291,13 +250,16 @@ sub schema {
     return $self->{schema};
   }
   elsif (ref $schema eq 'HASH') {
-    $self->_register_document($schema, $schema->{id} ||= 'http://generated.json.validator.url#');
+    $schema->{id} ||= $self->_default_id($schema);
+    warn "[JSON::Validator] Schema from hash. id=$schema->{id}\n" if DEBUG;
+    $schema = $self->_register_document($schema, $schema->{id});
   }
   else {
-    $schema = $self->_load_schema($schema)->data;
+    $schema = Cwd::abs_path($schema) if -e $schema;
+    $schema = $self->_load_schema($schema);
   }
 
-  $self->{schema} = Mojo::JSON::Pointer->new($self->_resolve_schema($schema, $schema->{id}, {}));
+  $self->{schema} = $self->_resolve_schema($schema, $schema->data->{id});
   $self;
 }
 
@@ -338,6 +300,17 @@ sub validate {
   return $self->_validate($data, '', $schema);
 }
 
+sub _build_formats {
+  return {
+    'date-time' => \&_is_date_time,
+    'email'     => \&_is_email,
+    'hostname'  => VALIDATE_HOSTNAME ? \&Data::Validate::Domain::is_domain : \&_is_domain,
+    'ipv4'      => VALIDATE_IP ? \&Data::Validate::IP::is_ipv4 : \&_is_ipv4,
+    'ipv6'      => VALIDATE_IP ? \&Data::Validate::IP::is_ipv6 : \&_is_ipv6,
+    'uri'       => \&_is_uri,
+  };
+}
+
 sub _coerce_by_collection_format {
   my ($self, $schema, $data) = @_;
   my $format = $schema->{collectionFormat};
@@ -349,30 +322,42 @@ sub _coerce_by_collection_format {
 }
 
 sub _load_schema {
-  my ($self, $url) = @_;
+  my ($self, $url, $parent) = @_;
   my ($namespace, $scheme) = ("$url", "file");
   my $doc;
 
-  if ($namespace =~ m!^https?://!) {
+  if ($namespace =~ $HTTP_SCHEME_RE) {
     $url = Mojo::URL->new($url);
-    ($namespace, $scheme) = ($url->clone->fragment(undef)->port(undef)->to_string, $url->scheme);
+    ($namespace, $scheme) = ($url->clone->fragment(undef)->to_string, $url->scheme);
   }
   elsif ($namespace =~ m!^data://(.*)!) {
     $scheme = 'data';
   }
+  elsif ($parent and $parent =~ $HTTP_SCHEME_RE) {
+    $parent = Mojo::URL->new($parent);
+    $url =~ s!#.*!!;
+    $url = $parent->path($parent->path->merge($url)->canonicalize);
+    ($namespace, $scheme) = ($url->to_string, $url->scheme);
+  }
+  elsif ($parent) {
+    $url =~ s!#.*!!;
+    $url = File::Spec->catfile(File::Basename::dirname($parent), split '/', $url);
+    $namespace = Cwd::abs_path($url) || $url;
+  }
 
   # Make sure we create the correct namespace if not already done by Mojo::URL
   $namespace =~ s!#.*$!! if $namespace eq $url;
 
   return $self->{cached}{$namespace} if $self->{cached}{$namespace};
   return eval {
-    warn "[JSON::Validator] Loading schema from $url ($namespace)\n" if DEBUG;
+    warn "[JSON::Validator] Loading schema $url namespace=$namespace scheme=$scheme\n" if DEBUG;
     $doc
       = $scheme eq 'file' ? Mojo::Util::slurp($namespace)
       : $scheme eq 'data' ? $self->_load_schema_from_data($url, $namespace)
       :                     $self->_load_schema_from_url($url, $namespace);
     $self->_register_document($self->_load_schema_from_text($doc), $namespace);
   } || do {
+    $doc ||= '';
     die "Could not load document from $url: $@ ($doc)" if DEBUG;
     die "Could not load document from $url: $@";
   };
@@ -399,57 +384,78 @@ sub _load_schema_from_url {
   return $doc;
 }
 
+sub _default_id {
+  my $path = Cwd::abs_path($0);
+  state $id = 0;
+  $path = File::Basename::dirname($path) if $path;
+  $path = Cwd::getcwd unless $path;
+  return File::Spec->catfile($path, sprintf 'json-validator-%s.json', ++$id);
+}
+
 sub _register_document {
   my ($self, $doc, $namespace) = @_;
 
   $doc = Mojo::JSON::Pointer->new($doc);
   $namespace = Mojo::URL->new($namespace) unless ref $namespace;
-  $namespace->fragment(undef)->port(undef);
-
-  warn "[JSON::Validator] Register $namespace\n" if DEBUG;
+  $namespace->fragment(undef);
 
   $self->{cached}{$namespace} = $doc;
   $doc->data->{id} ||= "$namespace";
   $self->{cached}{$doc->data->{id}} = $doc;
-  $doc;
-}
 
-sub _resolve_ref {
-  my ($self, $ref, $namespace, $refs) = @_;
-
-  return if !$ref or ref $ref;
-  $ref = "#/definitions/$ref" if $ref =~ /^\w+$/;
-  $ref = Mojo::URL->new($namespace)->fragment($ref) if $ref =~ s!^\#!!;
-  $ref = Mojo::URL->new($ref) unless UNIVERSAL::isa($ref, 'Mojo::URL');
-
-  return $refs->{$ref} if $refs->{$ref};
-
-  warn "[JSON::Validator] Resolve $ref\n" if DEBUG;
-  $refs->{$ref} = {};
-  my $doc = $self->_load_schema($ref);
-  my $def = $self->_resolve_schema($doc->get($ref->fragment), $doc->data->{id}, $refs);
-  delete $def->{id};
-  $refs->{$ref}{$_} = $def->{$_} for keys %$def;
-  $refs->{$ref};
+  warn "[JSON::Validator] Register id=$doc->{data}{id} namespace=$namespace\n" if DEBUG;
+  return $doc;
 }
 
 sub _resolve_schema {
-  my ($self, $obj, $namespace, $refs) = @_;
-  my $copy = ref $obj eq 'ARRAY' ? [] : {};
-  my $ref;
+  my ($self, $schema, $namespace) = @_;
+  my (@items, @refs);
 
-  if (ref $obj eq 'HASH') {
-    $obj = $ref if $ref = $self->_resolve_ref($obj->{'$ref'}, $namespace, $refs);
-    $copy->{$_} = $self->_resolve_schema($obj->{$_}, $namespace, $refs) for keys %$obj;
-    delete $copy->{'$ref'};
-    return $copy;
-  }
-  elsif (ref $obj eq 'ARRAY') {
-    $copy->[$_] = $self->_resolve_schema($obj->[$_], $namespace, $refs) for 0 .. @$obj - 1;
-    return $copy;
+  return $self->{resolved}{$namespace} if $self->{resolved}{$namespace};
+
+  warn "[JSON::Validator] Resolving schema $namespace\n" if DEBUG;
+  $self->{resolved}{$namespace} = Mojo::JSON::Pointer->new({%{$schema->data}});
+  @items = ($self->{resolved}{$namespace}->data);
+
+  # First step: Make copy and find $ref
+  while (@items) {
+    my $topic = shift @items;
+    if (ref $topic eq 'HASH') {
+      while (my ($k, $v) = each %$topic) {
+        $topic->{$k} = [@$v] if ref $v eq 'ARRAY';
+        $topic->{$k} = {%$v} if ref $v eq 'HASH';
+        push @refs, $topic if $k eq '$ref' and !ref $v;
+        push @items, $topic->{$k};
+      }
+    }
+    elsif (ref $topic eq 'ARRAY') {
+      push @items, @$topic;
+    }
   }
 
-  return $obj;
+  # Seconds step: Resolve $ref
+  for my $topic (@refs) {
+    my $ref = $topic->{'$ref'} or next;    # already resolved?
+    $ref = "#/definitions/$ref" if $ref =~ /^\w+$/;                         # TODO: Figure out if this could be removed
+    $ref = Mojo::URL->new($namespace)->fragment($ref) if $ref =~ s!^\#!!;
+    $ref = Mojo::URL->new($ref) unless ref $ref;
+
+    warn "[JSON::Validator] Resolving ref $ref defined in $namespace\n" if DEBUG == 2;
+    my $look_in = $self->{resolved}{$ref->clone->fragment(undef)};
+
+    if (!$look_in) {
+      $look_in = $self->_load_schema($ref, $namespace);
+      $look_in = $self->_resolve_schema($look_in, $look_in->data->{id} || $namespace);
+      warn "[JSON::Validator] Will look for $ref in $look_in->{data}{id}\n" if DEBUG == 2;
+    }
+
+    $ref = $look_in->get($ref->fragment || '')
+      || die qq[Could not find "$topic->{'$ref'}" ($ref). Typo in schema "$namespace"?];
+    %$topic = %$ref;
+    delete $topic->{id} unless ref $topic->{id};    # TODO: Is this correct?
+  }
+
+  return $self->{resolved}{$namespace};
 }
 
 sub _validate {
@@ -586,9 +592,6 @@ sub _validate_type_enum {
   return E $path, "Not in enum list: @$enum.";
 }
 
-# TODO: Need to figure out if this is a Swagger specific thing
-sub _validate_type_file { }
-
 sub _validate_format {
   my ($self, $value, $path, $schema) = @_;
   my $code = $self->formats->{$schema->{format}};
@@ -859,10 +862,8 @@ sub _guessed_right {
   return _guess_data_type($_[0]) eq $_[1] ? $_[1] : undef;
 }
 
-sub _is_byte_string { $_[0] =~ /^[A-Za-z0-9\+\/\=]+$/; }
-sub _is_date        { $_[0] =~ qr/^(\d+)-(\d+)-(\d+)$/io; }
-sub _is_date_time   { $_[0] =~ qr/^(\d+)-(\d+)-(\d+)T(\d+):(\d+):(\d+(?:\.\d+)?)(?:Z|([+-])(\d+):(\d+))?$/io; }
-sub _is_domain      { warn "Data::Validate::Domain is not installed"; return; }
+sub _is_date_time { $_[0] =~ qr/^(\d+)-(\d+)-(\d+)T(\d+):(\d+):(\d+(?:\.\d+)?)(?:Z|([+-])(\d+):(\d+))?$/io; }
+sub _is_domain { warn "Data::Validate::Domain is not installed"; return; }
 
 sub _is_email {
   state $email_rfc5322_re = do {
@@ -886,11 +887,6 @@ sub _is_ipv4 {
 
 sub _is_ipv6 { warn "Data::Validate::IP is not installed"; return; }
 
-sub _is_number {
-  return unless $_[0] =~ /^-?\d+(\.\d+)?$/;
-  return $_[0] eq unpack $_[1], pack $_[1], $_[0];
-}
-
 sub _is_true {
   return $_[0] if ref $_[0] and !Scalar::Util::blessed($_[0]);
   return 0 if !$_[0] or $_[0] =~ /^(n|false|off)/i;
